import copy
from django.contrib import admin
from django.utils.html import format_html



class ImageLiteAdmin(admin.ModelAdmin):
    '''
    Admin devised for maintenence of ImageLite models.
    Ot's more informative than stock.
    '''
    #! All the below can be commented or adapted to remove/change 
    # particular effects.
    # See the notes.
    #
    # case-insensitive 'contains' match
    search_fields = ['src']
    
    # Style the lists.
    # See the support code below
    list_display = ('filename', 'upload_day', 'image_delete', 'image_view',)
        
    # Support code for styling the admin list
    # See the 'list_display' attribute.
    def image_view(self, obj):
        return format_html('<a href="{}" class="button">View</a>',
            obj.src.url
        )
    image_view.short_description = 'View'

    def image_delete(self, obj):
        return format_html('<a href="{}/delete" class="button" style="background: #ba2121;color: #fff;">Delete</a>',
            obj.pk
        )
    image_delete.short_description = 'Delete'
    
    def upload_day(self, obj):
        '''e.g.	17 May 2020'''
        return format_html('{}',
            obj.upload_time.strftime("%d %b %Y")
        )
    upload_day.short_description = 'Upload day'
    upload_day.admin_order_field = '_upload_time'

    def filename(self, obj):
        return obj.filename
    filename.admin_order_field = 'src'
    
    # if you want to change the filechooser on the add form, you
    # can do it here. The default is
    # admin.widgets.AdminFileWidget
    #formfield_overrides = {
        ## For example, Drag and drop Image picker from,
        ## https://github.com/rcrowther/DDFileChooser
        #ImageFileField: {'widget': DDFileChooser},
    #}        



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _
from django.core import checks
from image_lite.checks import check_filters


class ImageLiteConfig(AppConfig):
    # NB It would be custom to test for operation of depend libraries
    # here, primarily Pillow. However, Django now does this for 
    # ImageFile. Also, the Wand files are a boxed and optional import. 
    # So not a concern.
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'image_lite'
    verbose_name = _("Image handling")
    

    def ready(self):
        super().ready()
        self.module.autodiscover()   
        checks.register(check_filters, 'image_filters')



from PIL import Image as PILImage
import math



def resize_aspect(pillow, width, height):
    '''
    Resize if the image is too big, preserving aspect ratio.
    Will not resize if dimensions match or are less than the source.
    @return an image within the given dimensions. The image is usually 
    smaller in one dimension than the given space.
    '''
    s = pillow.size
    current_width = s[0]
    current_height = s[1]
    width_reduce = current_width - width
    height_reduce = current_height - height
    if (width_reduce > height_reduce and width_reduce > 0):
        h =  math.floor((width * current_height)/current_width)
        return pillow.resize((width, h))
        
    #NB the equality. On the not unlikely chance that the width 
    # reduction is the same as the height reduction (for example, 
    # squares), reduce by height.
    elif (height_reduce >= width_reduce and height_reduce > 0):
        w =  math.floor((height * current_width)/current_height)
        return pillow.resize((w, height))
    else:
        return pillow



def crop(pillow, width, height):
    '''
    Crop if image is too big.
    Crop is anchored at centre. 
    Will not crop if both dimensions match or are less than the source.
    @return an image within the given dimensions. The image may be 
    smaller in one dimension than the given space.
    '''
    s = pillow.size
    current_width = s[0]
    current_height = s[1]
    
    width_reduce = current_width - width
    height_reduce = current_height - height

    # Only crop if the image is too big
    if ((width_reduce > 0) or (height_reduce > 0)):
        x = 0
        y = 0
        if (width_reduce > 0):
            x = width_reduce >> 1
        if (height_reduce > 0):
            y = height_reduce >> 1
        # Crop!
        pillow = pillow.crop((x, y, x + width, y + height))
    return pillow



def fill(pillow, width, height, fill_color="white"):
    '''
    Fill round an image to a box.
    Image must be smaller than the giveen box. Checking is 
    regarded as a seperate operation.
    '''
    # NB: This converts to RGB. Not an issue, as fill runs last.
    s = pillow.size
    current_width = s[0]
    current_height = s[1]

    x = (width - current_width) >> 1
    y = (height - current_height) >> 1
        
    bg = PILImage.new('RGB', (width, height), fill_color)

    # paste down.
    # I've tried a couple of ways, this mess seems to work ok
    if ((pillow.mode in ('RGBA', 'LA'))): 
        # The image has an alpha layer. But this may well be black, for
        # transparency, and on convert to RGB goes black. Solution,
        # split alpha layer and paste as mask.
        bg.paste(pillow, (x, y),  mask=pillow.getchannel('A'))
    else:
        # The image can be pasted on
        bg.paste(pillow, (x, y))
    return bg




from django.utils.functional import cached_property


'''
Map of allowed file extensions -> format key
Lower case entries, and variations in spelling 
e.g. 'jpg'/'jpeg'
'''    
# This needs to be a union of the filehandling of the possible 
# libraries, largely Pillow Wand, OpenCV
# Pillow
#   django.core.validators.get_available_image_extensions
# Imagemagic
#   > identify -list format
# (no Wand implementation?)
# I think we assume if Pillow can, Wand can. Also, we stay conservative.
# Pillow extensions: bmp, dib, gif, tif, tiff, 
# jfif, jpe, jpg, jpeg, pbm, pgm, ppm, pnm, png, apng, blp, bufr, cur, 
# pcx, dcx, dds, ps, eps, fit, fits, fli, flc, ftc, ftu, gbr, grib, h5, 
# hdf, jp2, j2k, jpc, jpf, jpx, j2c, icns, ico, im, iim, mpg, mpeg, 
# mpo, msp, palm, pcd, pdf, pxr, psd, bw, rgb, rgba, sgi, ras, tga, icb,
# vda, vst, webp, wmf, emf, xbm, xpm.
# Map possible (lowercase) file extensions -> internal format
# Avoid cross-media such as pdf, ps, mpg, even if libraries can handle.
EXTENSION_TO_APP_FORMAT = {
    'bmp' : 'bmp',
    'jpg' : 'jpg',
    'jpeg' : 'jpg',
    'png' : 'png',
    'gif' : 'gif',
    'ico' : 'ico',
    'rgb' : 'rgb',
    'rgba' : 'rgb',
    'tiff': 'tiff',
    'tif': 'tiff',
    'webp' : 'webp',
    'xbm' : 'xbm',
    'xpm' : 'xpm',
}

'''
Formats accepted.
This list is used throughout the app, and is definitive. The entries can
be used for new file extensions. However, for verification and other 
code, other collections in this module are preferred. The main use in 
code is for displaying nice messages for the user. 
'''
## To get extensions from internal format, use the internal format 
# string!
IMAGE_FORMATS = set()
for v in EXTENSION_TO_APP_FORMAT.values():
     IMAGE_FORMATS.add(v)

# def extensions_maxlen():
    # '''
    # max length of listed extensions, including the period.
    # eg '.tiff' returns 5
    # '''
    # m = 0
    # for f in IMAGE_FORMATS:
        # l = len(f)
        # if (l > m):
            # m = l
    # # for the period i.e '.png' is 4
    # m += 1
    # return m


'''
Map of uppercase format keys -> app formats.
This map will work for Pillow and Wand libraries.
'''
# Order matters. Put the reverse conversion last so it overwrites on 
# reverse generation
FORMAT_UCLIB_TO_APP = {
    'BMP' : 'bmp',
    'GIF' : 'gif',
    'ICO' : 'ico',
    'MPO' : 'jpg',
    'JPG' : 'jpg',
    'JPEG' : 'jpg',
    'PNG32' : 'png',
    'PNG64' : 'png',
    'PNG' : 'png',
    'RGB' : 'rgb',
    'RGBA' : 'rgb',
    'TIFF64' : 'tiff',
    'TIFF' : 'tiff',
    'WEBP' : 'webp',
}

'''
Map of app formats -> uppercase format keys.
This map will work for Pillow and Wand libraries.
'''
FORMAT_APP_TO_UCLIB = {}
for k,v in FORMAT_UCLIB_TO_APP.items():
    FORMAT_APP_TO_UCLIB[v] = k
    



import datetime

from django.core.management.base import BaseCommand, CommandError
from image_lite.management.commands import common
from pathlib import Path



class Command(BaseCommand):
    help = 'List reforms.'
    
    def add_arguments(self, parser):
        common.add_model_argument(parser)
        common.add_contains_argument(parser)
                
    def handle(self, *args, **options):
        reform_path = common.get_reform_path(options)
        reform_list = [fp.stem for fp in reform_path.iterdir() if fp.is_file()]
        contains_test = options["contains"]
        if (contains_test):
            reform_list = filter(lambda fn: contains_test in fn, reform_list)
        rl = sorted(reform_list)
        for fp in rl:
            print(fp)



import math
from django.utils.functional import cached_property
from django.templatetags.static import static


def app_name(klass):
    '''
    Get an app name from a klass
        return the appname. IF the name is empty, it returns empty
    '''
    # appears to be the most complete way of getting an appname.
    # Doesn't rely on, for example, the class being a package
    r = klass.__module__.split('.', 1)[0]
    if (r == ""):
        raise Exception(f"App name requested but not found. klass:{klass}")
    return r
            
def bytes2mb(v):
    # This is how Django does it, I think, binary MB
    #NB must handle fractions so not shifts
    return v / 1048576

def mb2bytes(v):
    # This is how Django does it, I think, binary MB
    #NB must handle fractions so not shifts
    return math.ceil(v * 1048576)

def url_absolute_static_aware(path):
    """
    Given a relative or absolute path to a static asset, return an 
    absolute path. 
    
    return
        An absolute path will be returned unchanged while a 
        relative path will be passed to 
        django.templatetags.static.static(). This prefixes with 
        STATIC_URL or a staticfiles_storage.storage defined prefix 
        (usually '/static/').
    """
    # Straight lift from django.forms.widgets.Media
    if path.startswith(('http://', 'https://', '/')):
        return path
    return static(path)

#! Where, and for what?
# from django.utils.encoding import iri_to_uri
# from urllib.parse import quote, urljoin
# from django.conf import settings


#! should match names with pathlib.Path
# class ModulePath():
    # '''
    # Path-like class for handling import/module patths.
    # '''
    # @classmethod
    # def from_str(class_object, str_path):
        # '''Call as
           # d = ModulePath.from_str('graphic.effect.zoom')
        # '''
        # return class_object(*str_path.split('.'))
        
    # def __init__(self, *args):
        # self.path = args

    # @property
    # def size(self):
        # return len(self.path)
        
    # @property
    # def leaf(self):
        # return self.path[-1]
               
    # @property
    # def root(self):
        # return self.path[0]
        
    # @property
    # def branch(self):
        # if (self.size > 1):
            # return ModulePath(*self.path[0:-1])
        # else:
            # raise IndexError('Can not return branch when length is one. elem:{}'.format(root))
            
    # def extend(self, new_leaf):
        # return ModulePath(*self.path, new_leaf)
        
    # @property
    # def str(self):
        # return self.__str__()
        
    # def __repr__(self):
        # return 'ModulePath("{}")'.format('", "'.join(self.path))
        
    # def __str__(self):
        # return ".".join(self.path)
        




from pathlib import Path
import os
from django.core.management.base import BaseCommand, CommandError
from django.core.files.images import ImageFile
from image_lite.management.commands import common
from image_lite import registry



class Command(BaseCommand):
    help = 'Automatically/bulk generate reforms to an image model. The command tries to ignore errors and continue. It will append to existing collections. Attributes other than file data will default.'

    def add_arguments(self, parser):
        common.add_model_argument(parser)
        # parser.add_argument(
            # 'src_path', 
            # type=str
        # )


    #! accept liat of filter names
    def handle(self, *args, **options):
        Model = common.get_image_model(options)
        #- get storage
        storage = common.get_storage(Model)
        
        # Get full reform path
        reform_path_from_media = common.get_reform_path(options)
        reform_base_path = Path(storage.location) / reform_path_from_media

        # Need to...
        #- get filters, 
        filters = registry(Model)

        # get filenames DB holds
        images_qs = Model.objects.all() #values_list('src', flat=True)

        count = 0
        ignored = 0
        fail = []
        for image in images_qs:
            # construct a basic reform path
            src_filename = Path(image.src.name).stem
            reform_file_path = Path(reform_base_path) / src_filename
            #print(str(reform_file_path))

            # We've only got started....
            for filter_class in filters:
                reform_path = Path(
                    filter_class.add_suffix_to_local_path(reform_file_path)
                )
                
                # - bail if file exists 
                if (reform_path.exists()):
                    ignored += 1
                    continue
         
                # - run filter on existing files
                with image.src.open() as fsrc:
                    filter_instance = filter_class()
                    (reform_buff, iformat) = filter_instance.process(
                        fsrc,
                        #model_args
                        {}
                    )

                # save
                with open(reform_path, "wb") as f:
                    try:
                        f.write(reform_buff.getbuffer()) 
                        count += 1
                    except Exception:
                        fail.append(reform_path)
                    
        # output some results            
        if (options['verbosity'] > 0):
            print(f"{count} image(s) created") 
            print(f"{ignored} reform images ignored because they exist")
            if (len(fail) > 0):
                print("{} image(s) failed save. Basenames: '{}'".format(
                    len(fail),
                    "', '".join(fail),
                    )) 




from image_lite import checks



class PlacementError(Exception):
    """
    Custom exception for attempted registration of Filters not in an 'image_filers' module
    """
    pass


        
class Filter():
    '''
    A filter defines how to transorm a image with a spec into 
    a derivative with a different spec.
    As such, it is more than a static configuration. It is half-
    functional. takes an open Python file and returns a buffer. 
    The stock implementation uses Pillow, but this setup allows 
    other codebases to bue used in a filler without changing calling 
    code.
    '''
    #? cache
    @classmethod
    def _module_path_human(cls):
        ''' 
        Human readable module path.
        Used to refer to a filter in the registry and templates. 
        Currently based on the module path and classname.
        '''
        mp = cls.__module__.split('.')
        
        # We encourage users to place filters in a 
        # module called image_filters'. This would involve a lot of
        # repetition in template code. So an 'image_filters' element is removed from the path.
        # That risks a collision with
        # filters placed in the root of an app, so insist on placement in 'image filters'.
        if (mp[-1] == 'image_filters'):
            del(mp[-1])
        else:
            raise PlacementError("Please locate Filters to be registered in an 'image filters' module: {}.{}".format(
                cls.__module__,
                cls.__name__
            ))        
        return mp
        
    @classmethod
    def name(cls):
        return cls.__name__
        
    @classmethod
    def human_id(cls):
        ''' 
        id for the filter.
        Used to refer in the registry and templates, so is human 
        readable. 
        Currently based on a dotted module path and classname.
        '''
        p = cls._module_path_human()
        p.append( cls.__name__ )
        
        # That is unique, but we encourage users to place filters in a 
        # module called image_filters'. This would involve a lot of
        # repetition in template code. At the risk of a collision with
        # filters placed in the root of an app, we remove that part of 
        # the path.
        return ".".join(p)
        
    #x unused?
    def filename(self, stem, extension):
        '''
        Generate a filename from a base string.
        An ammended filter path and name is appended to the stem,
        
        stem + '-' + ammended underscore path + '_' + filtername.lower() + '.' + extension
         
        It is to be concise, and near-unique to avoid provoking storage backends, 
        '''
        return "{}-{}_{}.{}".format(
            stem,
            "_".join(self._module_path_human()),
            type(self).__name__.lower(),
            extension
        )
        
    @classmethod
    def check(cls, **kwargs):
        return []

    def process(self, src_file, model_args):
        '''
        Wrap a Python file handle for processing a reform.
        Return should be a BytesIO buffer or similar. Some APIs do not
        include file saving, but most will deal with a generic Python
        buffer.
                 
        'save_info_callback' makes decisions about how the resulting 
        buffer should be saved. There is no need to call this function 
        for dev, but final versions should respect the environment by 
        calling it and acting on the results. Current implementation 
        needs the source format, then checks with overall and filter 
        settings, then returns decisions on 'format' and 'jpeg_quality'.
                
        src_file 
            an open Python file handle 
        return 
            a BytesIO or similar
        '''
        raise NotImplementedError
        
        
        
# Classes here on are mixins. They estabish attributes and checks, 
# skeletons to hang image-processing code on.
class FormatMixin():
    '''Establish the format for an image. 
    Set format=None means the image is unchanged.
    Most filters will contain and build from this base class.
    '''
    format=None
    jpeg_quality=None
    
    @classmethod
    def add_suffix_to_path(cls, base_path):
        return str(base_path) + '-' + cls.__name__ + '.' + cls.format
        
    @classmethod
    def check(cls, **kwargs):
        errors = super().check(**kwargs)
        errors += [
            *checks.check_image_format_or_none('format', cls.format, 'image_filter.E001', **kwargs),
            *checks.check_jpeg_has_quality( cls.format, cls.jpeg_quality, 'image_filter.E002', **kwargs),
            *checks.check_jpeg_quality_range(cls.jpeg_quality, 'image_filter.E003', **kwargs),
            *checks.check_jpeg_legible(cls.jpeg_quality, 'image_reform.W001', **kwargs),
        ]
        return errors


class PhotoFXMixin():
    '''Establish the format for an image. 
    Set format=None means the image is unchanged.
    Most filters will contain and build from this base class.
    '''
    pop=False
    greyscale=False    
    night=False
    warm=False
    strong=False
    film=False
    no=False
    watermark=''
    
    @classmethod
    def check(cls, **kwargs):
        errors = super().check(**kwargs)
        errors += [
            *checks.check_boolean('pop', cls.pop, 'image_filter.E003', **kwargs),
            *checks.check_boolean('greyscale', cls.greyscale, 'image_filter.E004', **kwargs),
            *checks.check_boolean('night', cls.night, 'image_filter.E005', **kwargs),
            *checks.check_boolean('warm', cls.warm, 'image_filter.E006', **kwargs),
            *checks.check_boolean('strong', cls.strong, 'image_filter.E007', **kwargs),
            *checks.check_boolean('film', cls.film, 'image_filter.E008', **kwargs),
            *checks.check_boolean('no', cls.no, 'image_filter.E009', **kwargs),
            #NB Would test for watermark existance, but it is URL access,
            # thus server dependant.
        ]
        return errors
        
                           
                                
class ResizeCropMixin():
    '''Resize n image.
    Shrinks inside the box defined by the given args. So the result will
    usually be smaller width or height than the given box.
    A base class.
    '''
    width = None
    height = None
    
    @classmethod
    def check(cls, **kwargs):
        errors = super().check(**kwargs)
        errors += [
            *checks.check_positive('width', cls.width, 'image_filter.E011', **kwargs),
            *checks.check_positive('height', cls.height, 'image_filter.E012', **kwargs),
        ]
        return errors
        
                                    
class ResizeCropSmartMixin(ResizeCropMixin):
    '''Resize an image.
    This resize lays the image on a background of ''fill-color'.
    So the result always matches the given sizes.
    A base class
    '''
    fill_color="white"

    @classmethod
    def check(cls, **kwargs):
        errors = super().check(**kwargs)
        errors += [
        #? No test for color
        ]
        return errors



from django.core import checks
from image_lite.constants import IMAGE_FORMATS



# Needs clearout
## This app speccific
def check_filters(**kwargs):
    '''
    Check registered filters with Django static checks framework.
    '''
    from image_lite import registry
    errors = []
    #print(str(registry))
    for filter_class in registry.list_entries():
        #print(str(filter_class))
        filter_instance = filter_class()
        errors.extend(filter_instance.check())
    return errors


def _image_format_error(setting_name, v, eid, accepted_formats):
    return checks.Error(
                "'{}' format '{}' unrecognised."
                " Recognised image formats: {}".format(
                setting_name,
                v,
                ", ".join(accepted_formats),
            ),
            id=eid,
    )
    
def check_image_format_or_none(
        setting_name,
        image_format,
        eid, 
        accepted_formats=IMAGE_FORMATS,
        **kwargs
    ):
    '''
    Accepts a format
    '''
    errors = []
    if (image_format is None):
        return errors
    if (not(image_format in accepted_formats)):
        errors.append(_image_format_error(setting_name, v, eid, accepted_formats))
    return errors 
    
def check_image_formats_or_none(
        setting_name, 
        image_formats,
        eid, 
        accepted_formats=IMAGE_FORMATS,
        **kwargs
    ):
    '''
    Accepts a list of formats
    '''
    errors = []
    if (image_formats is None):
        return errors
    unrecognised_formats = [f for f in image_formats if not(f in accepted_formats)]
    if (unrecognised_formats):
        v = ", ".join(unrecognised_formats)
        errors.append(_image_format_error(setting_name, v, eid, accepted_formats))
    return errors 
    
def check_jpeg_has_quality(img_format, jpeg_quality, eid, **kwargs):
    errors = []
    if (img_format == 'jpg' and (not jpeg_quality)):
        errors.append(
            checks.Error(
                "If format is 'jpg' then 'jpeg_quality' must be declared.",
                id=eid,
            ))   
    return errors  
    
def check_jpeg_quality_range(jpeg_quality, eid, **kwargs):
    errors = []
    if (jpeg_quality and (jpeg_quality < 1 or jpeg_quality > 100)):
        errors.append(
            checks.Error(
                "'jpeg_quality' value '{}' must be 1--100.".format(
                jpeg_quality
            ),
            id=eid,
        ))   
    return errors  

def check_jpeg_legible(jpeg_quality, eid, **kwargs):
    errors = []
    if (jpeg_quality and (jpeg_quality < 12)):
        errors.append(
            checks.Warning(
                "'jpeg_quality' value '{}' is very low.".format(
                jpeg_quality
            ),
            id=eid,
        ))   
    return errors 



## General
def check_is_subclass(setting_name, v, base_klass, eid, **kwargs):
    errors = []
    if (not(issubclass(v, base_klass))):
        errors.append(
            checks.Error(
                "''{}' value {} must be a subclass of {}.".format(
                setting_name,
                v,
                base_klass.__name__,
            ),
            id=eid,
        ))
    return errors
    
def check_type(setting_name, v, tpe, eid, **kwargs):
    errors = []
    if (not(type(v)==tpe)):
        errors.append(
            checks.Error(
                "'{}' value '{}' must be type {}.".format(
                setting_name, 
                v,
                tpe
            ),
            id=eid,
        ))
    return errors 

def check_boolean(setting_name, v, eid, **kwargs):
    return check_type(setting_name, v, bool, eid, **kwargs)

def check_int(setting_name, v, eid, **kwargs):
    return check_type(setting_name, v, int, eid, **kwargs)
        
def check_positive(setting_name, v, eid, **kwargs):
    errors = check_int(setting_name, v,  eid, **kwargs)
    if ((not errors) and int(v) <= 0):
        errors.append(
            checks.Error(
            "'{}' value '{}' must be a positive number.".format(
            setting_name, 
            v
            ),
            id=eid,
        ))
    return errors 

def check_positive_float_or_none(setting_name, v, eid, **kwargs):
    errors = []
    if (v is None):
        return errors
    try:
        if (float(v) <= 0):
            raise TypeError
    except TypeError:    
        errors.append(
            checks.Error(
            "'{}' value '{}' must be None or a positive float.".format(
            setting_name, 
            v
            ),
            id=eid,
        ))
    return errors
             
def check_numeric_range(setting_name, v, imin, imax, eid, **kwargs):
    errors = []
    try:
        if (type(v) != int or v < imin or v > imax):
            raise TypeError
    except TypeError: 
        errors.append(
            checks.Error(
                "'{}' value '{}' must be > {} and < {}.".format(
                setting_name, 
                v,
                imin,
                imax,
                ),
                id=eid,
        ))
    return errors

# def check_file_exists(setting_name, v, eid, **kwargs):    
    # errors = []
    # if (v and (not(Path(v).is_file()))):
        # errors.append(
            # checks.Error(
            # "'{}' value '{}' can not be deetected as an existing file".format(
            # setting_name, 
            # v
            # ),
            # id=eid,
        # ))
    # return errors 

    # def _check_available_filelength(cls, **kwargs):
        # '''
        # Does filepath_length allow for filenames?
        # '''
        # errors = []
        # print(str(cls.filepath_length))
        # declared_len = int(cls.filepath_length)
        # path_len =  len(cls.upload_dir)
        # if (declared_len <= path_len):
            # errors.append(
                # Error(
                    # "'filepath_length' must exceed base path length. 'filepath_length' len: {}, 'upload_dir' len: {}".format(
                     # declared_len,
                     # path_len,
                     # ),
                    # id='image_model.E002',
                # )
            # )
        # elif (declared_len <= (path_len + 12)):
            # errors.append(
                # Warning(
                    # "Less than 12 chars avaiable for filenames. 'filepath_length' len: {}, 'upload_dir' len: {}".format(
                     # declared_len,
                     # path_len,
                     # ),
                    # id='image_model.W001',
                # )
            # )
        # return errors



from contextlib import contextmanager
from pathlib import Path
from django.db import models, transaction
from django.core.checks import Error, Warning
#from django.urls import reverse
from django.core.files.images import ImageFile
from django.apps import apps
from django.conf import settings
from django.utils.translation import gettext_lazy as _
from django.utils.functional import cached_property
from image_lite import checks
from image_lite.model_fields import ImageLiteField
from image_lite import registry



class SourceImageIOError(IOError):
    """
    Custom exception to distinguish IOErrors that were thrown while opening the source image
    """
    pass
    

def get_image_upload_to(instance, filename):
    """
    Obtain a valid upload path for an image file.
    This needs to be a module-level function so that it can be 
    referenced within migrations, but simply delegates to the 
    `get_upload_to` method of the instance, so that AbstractImage
    subclasses can override it.
    """
    return instance.get_upload_to(filename)
    


class AbstractImage(models.Model):
    '''
    Data about stored images.
    
    Provides db fields for width, height, bytesize and upload_data. 
    These replace storage backend ingenuity with recorded data.
    Also provides accessors for useful derivatives such as 'alt' and 
    'url'.
    
    Handles upload_to in a configurable way, and also provides 
    machinery for Reform handling.
    
    A note on configuration. An Image or Reform is always expected to 
    point at a file, it is never null. To not point at a file is an 
    error---see shortcuts and 'broken image'. Whatever model keys an 
    Image/Reform is still free to be null.
    
    An Images/Reform model is locked to a file folder. New models, even 
    if given files from the same folder, are file-renamed by Django. 
    Thus each file is unique, and each file field in the model is 
    unique.  
    '''
    # relative to MEDIA_ROOT
    upload_dir='originals'

    # 100 is Django default
    # Same as max_length on the src field, but this setting 
    # overrides that. This setting is checked to be between 1 and 
    # sixteen bits. Nut bear in mind length represents codepoints, so 
    # 20,000 codepoints is a large field for the database to store.
    # PS. Win32 operating systems can handle 255 char path lengths
    filepath_length=100
    
    # limit the unload filename length by checking on generated forms.
    # (if false, all filenames are accepted then truncated if necessary)
    form_limit_filepath_length=True
    
    # List of formats accepted. Should be lower-case, short form.
    # If None, any format recognised by this app as an image.
    accept_formats = None
    
    # If None, any size allowed. In MB. Real numbers allowed.
    max_upload_size = 2
    
    #  on DB record deletion, delete originals.
    auto_delete_files=False
    
    ## Reform options ##
    # Reforms naturally inherit filename options
    reform_dir='reforms'
    
    
    def delete_file_and_reforms(self):
        '''
        Delete files, original and reforms, associated with this model.
        Reform removal is by constructed paths, so reasonably efficient,
        but will not repair changed filters. Failure to delete reforms
        is ignored.
        '''
        # Delete reforms
        cls = self.__class__
        reform_dir_path = cls.reform_dir_path() / self.filename
        for filter_class in cls.filters():
            reform_path = Path(filter_class.add_suffix_to_path(reform_dir_path))
            reform_path.unlink(missing_ok=True)
            
        # delete original
        self.src.delete(False)            
            
    @classmethod
    def delete_file(cls, instance, **kwargs):
        transaction.on_commit(lambda: instance.delete_file_and_reforms())

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__()
        if (cls.auto_delete_files):
            models.signals.post_delete.connect(cls.delete_file, sender=cls)
          
    upload_time = models.DateTimeField(_("Datetime of upload"),
        auto_now_add=True, editable=False
    )

    # A note about the name. Even if possible, using the word 'file'
    # triggers my, and probably other, IDEs. Again, even if possible,
    # naming this the same as the model is not a good idea, if only due 
    # to confusion in relations, let alone stray attribute manipulation 
    # in Python code. So, like HTML, it is 'src'     
    src = ImageLiteField(_('image_file'), 
        #storage = FileSystemStorage,
        unique=True,
        upload_to=get_image_upload_to, 
        width_field='width', 
        height_field='height',
        bytesize_field="bytesize",
    )
    
    # Django can use Pillow to provide width and height. So why?
    # The 'orrible duplication, which Django supports, is to spare web 
    # hits on remote storage, and opening and closing the file for data. 
    # Sure, the values could be cached, but here assuming code needs
    # a solid record.
    width = models.PositiveIntegerField(verbose_name=_('width'), editable=False)
    height = models.PositiveIntegerField(verbose_name=_('height'), editable=False)

    # See the property
    bytesize = models.PositiveIntegerField(null=True, editable=False)

    def is_local(self):
        """
        Is the image on a local filesystem?
        return
            True if the image is hosted on the local filesystem
        """
        try:
            self.src.path
            return True
        except ValueError as e:
            # The access attempt will fail with currently a ValueError
            # if no associated file. But in context of asking 'is it 
            # local?' that's now a source error.
            raise SourceImageIOError(str(e))
        except NotImplementedError:
            # if not local storage, return false
            return False

    def get_upload_to(self, filename):
        # Incoming filename comes from upload machinery, and needs 
        # path appending.
        return Path(self.upload_dir) / filename
        
    @contextmanager
    def open_src(self):
        # Open file if it is closed
        close_src = False
        try:
            src = self.src

            if self.src.closed:
                # Reopen the file
                if self.is_local():
                    self.src.open('rb')
                else:
                    # Some external storage backends don't allow reopening
                    # the file. Get a fresh file instance. #1397
                    storage = self._meta.get_field('src').storage
                    src = storage.open(self.src.name, 'rb')

                close_src = True
        except IOError as e:
            # IOError comes from... an IO error. 
            # re-throw these as a SourceImageIOError
            # so that calling code
            # can distinguish these from IOErrors elsewhere in the 
            # process e.g. currently causes a broken-image display.
            raise SourceImageIOError(str(e))

        # Seek to beginning
        src.seek(0)
        try:
            yield src
        finally:
            if close_src:
                src.close()

    @classmethod
    def reform_dir_path(cls):
        location = cls._meta.get_field('src').storage.location  
        path = Path(location) / cls.reform_dir
        if (not(path.is_dir())):
            raise CommandError(f"Reform path not recognised as a directory. path: '{path}'")
        return path 
    
    @classmethod
    def filters(cls):
        '''
        Returns a list of Filter classes configured on this model.
        '''
        return registry(cls)

    @classmethod
    def get_filter(cls, filter_name):
        '''
        Returns a Filter class configured on this model, or None.
        '''
        class_list = registry(cls)
        r = None
        for k in class_list:
            if (k.__name__ == filter_name):
                r = k
                break
        return r
        
    def save(self, *args, **kwargs):
        # use save
        # Won't work for bulk creates, but niether will signals
        # https://docs.djangoproject.com/en/3.2/topics/db/models/#overriding-model-methods
        super().save(*args, **kwargs)

        # make a base path and filename
        reform_base_path = Path(settings.MEDIA_ROOT) / self.reform_dir
    
        ##? on app registration   
        #Path(dst_dir).mkdir(parents=True, exist_ok=True)

        fname = str(Path(self.filename).stem)
        reform_file_path =  Path(reform_base_path) / fname
        
        # run filters on file
        filters = registry(self)
        for filter_class in filters:
            reform_path = filter_class.add_suffix_to_path(reform_file_path)

            # get filtered buffer
            with self.src.open() as fsrc:
                filter_instance = filter_class()
                (reform_buff, iformat) = filter_instance.process(
                    fsrc,
                    #model_args
                    {}
                )
                
            with open(reform_path, "wb") as f:
                f.write(reform_buff.getbuffer())

    @property
    def filename(self):
        '''
        File as name, no path, no extension.
        Useful for admin displays, and others.
        '''
        return Path(self.src.name).stem

    # @property
    # def alt(self):
        # '''
        # String for an 'alt' field.
        # The base implementation is derived from the filepath of the 
        # uploaded file. 
        # Subclasses might override this attribute to use more refined 
        # data, such as a slug or title.
        # '''
        # return Path(self.src.name).stem + ' image'
                
    def is_portrait(self):
        return (self.width < self.height)

    def is_landscape(self):
        return (self.height < self.width)

    @classmethod
    def check(cls, **kwargs):
        errors = super().check(**kwargs)
        name = cls.__name__.lower()
        if not cls._meta.swapped:
            errors += [
            #NB By the time of check() models are built. So all 
            # attributes exist, at least as default.
            #*checks.check_type('reform_model', cls.reform_model, str, '{}.E001'.format(name), **kwargs),
            #*checks.check_str('upload_dir', cls.upload_dir, 1, '{}.E002'.format(name), **kwargs),
            *checks.check_numeric_range('filepath_length', cls.filepath_length, 1, 65535, '{}.E003'.format(name), **kwargs),
            *checks.check_image_formats_or_none('accept_formats', cls.accept_formats,'{}.E004'.format(name), **kwargs),
            *checks.check_positive_float_or_none('max_upload_size', cls.max_upload_size, '{}.E003'.format(name), **kwargs),
            ]
        return errors

    def __repr__(self):
        return "{}(upload_time: {}, src:'{}')".format(
            self.__class__.__name__,
            self.upload_time,
            self.src,
        )                

    def __str__(self):
        return self.src.name

    class Meta:
        abstract = True
        



from django.apps import apps
from django.core.exceptions import ImproperlyConfigured
from image_lite.models import AbstractImage


## These utils can only be loaded when Django/the app is ready
#? could use django.utils.module_loading.import_string()
def get_model(model_path, app_name=''):
    '''
    Get a model from a string path.
    
    model_path
         a strin path in form 'app_name.model_name'
    app_name
        an app name. If the model_path will only parse to a model, this 
        is used to provide the app name.
    return
        a model, or ImproperlyConfigured
    '''
    model_path_elements = model_path.split('.', 1)
    if (len(model_path_elements) == 1):
        if (not app_name):
            raise ImproperlyConfigured("Unable to parse model path (short name, no app data): '{}'".format(model_path))
        else:
            an = app_name            
            if callable(app_name):
                an = app_name()
            mn = model_path_elements[0]
    if (len(model_path_elements) == 2):
        an = model_path_elements[0]
        mn = model_path_elements[1]
    if (len(model_path_elements) > 2):
        raise ImproperlyConfigured("Unable to parse given model path: '{}'".format(model_path))
    try:
        Model = apps.get_model(an, mn)
    except Exception as e:
        raise ImproperlyConfigured("Unable to locate model from model path: '{}'".format(model_path))
    return Model

def get_image_model(model_path, app_name=''):
    '''
    Get an model from a string path.

    model_path
         a string path in form 'app_name.model_name'
    app_name
        an app name. If the model_path will only parse to a model, this 
        is used to provide the app name.
    return
        if path empty, Image, if path, an Image model, if path fails, 
        ImproperlyConfigured
    '''
    Model = get_model(model_path, app_name)
    if (not issubclass(Model, AbstractImage)):
        raise ImproperlyConfigured("Given class not a subclass of AbstractImage: '{}'".format(
            model_path
        ))
    return Model




from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from pathlib import Path


class ModelUniqueFilenameMixin:

    def validate_unique(self, exclude=None):
        filepath = Path(self.upload_dir) / self.src.name
        qs = self.__class__._default_manager.filter(src=filepath)
        if qs.exists():
            raise ValidationError({'src': _('Fileename exists in image repository')})
        super().validate_unique(exclude)



from collections import Iterable 
from image_lite import utils


class AlreadyRegistered(KeyError):
    pass


class NotRegistered(KeyError):
    pass
    
    
class Unregisterable(Exception):
    pass


class ClassByModuleRegistry:
    '''
    Simple registry.
    This is useful for locating code. Why do that when Python's import
    system is flexible and concise?
    Because you may wish to refer to code by string names. This is 
    relevant when code needs to be accessed from Django templates,
    where parameters are (mostly) strings.
    - Why not use a cached property?
    - Use decorators? 
    Because this allows targeted classes, so other code can be put in 
    the targetted modules.
    Does not handle permissions. 
    '''
    def __init__(self):
        # app_name -> list(class (not instance))
        self._registry = {} 



    def __contains__(self, klass):
        app_name = utils.app_name(klass)
        class_name = klass.__name__
        class_list = self._registry.get(app_name, [])
        r = False
        for k in class_list:
            r = r or (k.__name__ == class_name)
        return r
        
    def register(self, klass):
        #print(str(klass.__module__.split('.', 1)[0]))
        app_name = utils.app_name(klass)
        if (not app_name in self._registry):
            self._registry[app_name] = []
        else:
            class_list = self._registry[app_name]
            if klass in class_list:
                raise AlreadyRegistered('Already registered. class:{}'.format(
                    klass.__name__
                ))
        self._registry[app_name].append(klass)
                
    # def unregister(self, klass):
        # """
        # Unregister the given model(s).

        # If a model isn't already registered, raise NotRegistered.
        # """
        # if k not in self._registry:
            # raise NotRegistered('Class can not be unregistered {}'.format(k))
        # del self._registry[k]

    def __call__(self, klass_or_instance):
        '''
        A list of classes registered in the module.
        '''
        r = None
        app_name = utils.app_name(klass_or_instance) 
        try:
            r = self._registry[app_name]
        except KeyError:
            raise NotRegistered("Module entries requested but not found. app_name:{}".format(
            app_name
            ))
        return r
        
    def get_filter(self, app_name, filter_name):
        class_list = self._registry.get(app_name, [])
        r = None
        for k in class_list:
            if (k.__name__ == filter_name):
                r = k
                break
        return r
                
    @property
    def list_apps(self):
        return self._registry.keys()

    def list_entries(self):
        r = []
        for l in self._registry.values():
            r.extend(l)
        return r
        
    @property
    def size(self):
        return len(self._registry)    
        
    def __str__(self):
        r = []
        for app_name, class_list in self._registry.items():
            #class_nameB = []
            #print(str(class_list))
            #for klass in class_list:
            #    class_nameB.append(klass.__name__)
            class_nameB = [klass.__name__ for klass in class_list]
            class_names = ", ".join(class_nameB) 
            #print(str(nameList))
            r.append( f"{app_name}=>({class_names})" )
        class_list_by_app = ", ".join(r) 
        return f"{self.__class__.__name__}({class_list_by_app})"
        
from image_lite.filters import Filter

class FilterRegistry(ClassByModuleRegistry):
    def register(self, class_or_iterable):
        if (not isinstance(class_or_iterable, Iterable)):
            class_or_iterable = [class_or_iterable]
        for filter_class in class_or_iterable:
            if (not (issubclass(filter_class, Filter))):
                raise Unregisterable("Class is not a subclass of Filter. class:{}".format(
                    filter_class.__path__
                ))
            super().register(filter_class)

    # def unregister(self, class_or_iterable):
        # if (not isinstance(class_or_iterable, Iterable)):
            # class_or_iterable = [class_or_iterable]
        # for filter_class in class_or_iterable:
            # super().unregister(filter_class)
      
registry = FilterRegistry()



from django.db.models import (
    ImageField, 
    signals
)
from image_lite import utils
from image_lite import form_fields



class ImageLiteField(ImageField):
    '''
    A (model) ImageField.
    ImageLiteField is adapted from Imagefield 
    - length of file path is optional
    - acceptable formats can be added, limiting the apps more general 
    list
    - max upload size is optionally checked
    - Maintains a field of configurable name containing file bytesize
    On initialisation, the field sends a signal to itself, which
    triggers a file read for bytesize. The method 
    update_bytesize_field() is also available for external triggering 
    by, for example, a file update. 
    The formfield defaults to a TighterImageField, which can can skip 
    length verification, and precheck filesize and extensions.
    
    max_size
        in bytes
    '''
    #NB In stock Django,
    # - The model ImageField does routing, not validation, not even 
    # through FileField
    # - The form FileField read-checks files. It checks if a file exists, 
    # has a size, and that the filename is not too long.
    # - The form ImageField goes a little further. It checks Pillow can 
    # read the file, that Pillow.verify() does not think it is broken, 
    # that the MIME type is coherent, then validates the extension 
    # against Pillow data.
    # - Interesting, ImageFile itself has a little checking, as it must 
    # rescue dimension data from Pillow.

    # Class mainly exists for contribute_to options, faking abstract
    # classes. But adds some init vals, bytesize handling, and it's 
    # formfield belongs to this app
    default_validators = []
    
    def __init__(self, 
        verbose_name=None,
        name=None, 
        bytesize_field=None,
        accept_formats=None,
        form_limit_filepath_length = True,
        max_size=None,
        **kwargs
    ):
        self.bytesize_field = bytesize_field
        self.accept_formats = accept_formats
        self.form_limit_filepath_length = form_limit_filepath_length
        self.max_size = max_size
        self.default_validators = []
        super().__init__(verbose_name, name, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        # upload_to and max_length handled by super()
        if self.bytesize_field:
            kwargs['bytesize_field'] = self.bytesize_field           
        kwargs['max_size'] = self.max_size
        kwargs['form_limit_filepath_length'] = self.form_limit_filepath_length
        return name, path, args, kwargs

    def contribute_to_class(self, cls, name, private_only=False):
        # This is not contribute to class at all. It is class 
        # contribute to field. That said, Django does this for 
        # verbose names.
        # 
        # Point is, this method is run by the metaclass
        # way before the classes are constructed. And it is run on
        # every class in MRO, from base to subclasses. So class 
        # attributes referenced here can influence actions in an 
        # imitation of overriding. Which Python cannot otherwise do.  
        super().contribute_to_class(cls, name, private_only=False)

        # Since this predates init, and steps from base, we can simply
        # overwrite values
        # Must use hasattr() defence because migration builds 
        # '__fake__' classes
        if (hasattr(cls, 'filepath_length')):
            self.max_length = cls.filepath_length
        if (hasattr(cls, 'form_limit_filepath_length')):
            self.form_limit_filepath_length = cls.form_limit_filepath_length
        if (hasattr(cls, 'max_upload_size')):
            self.max_size = utils.mb2bytes(cls.max_upload_size)
        if (hasattr(cls, 'accept_formats')):
            self.accept_formats = cls.accept_formats

        # slightly more normal
        # signal to update named bytesize field
        if not cls._meta.abstract:
            signals.post_init.connect(self.update_bytesize_field, sender=cls)

    def update_bytesize_field(self, instance, force=False, *args, **kwargs):
        # Nothing to update if the field doesn't have the attribute or if
        # the field is deferred.
        if not self.bytesize_field or self.attname not in instance.__dict__:
            return

        # getattr will call the ImageFileDescriptor's __get__ method, which
        # coerces the assigned value into an instance of self.attr_class
        # (ImageFieldFile in this case).
        file = getattr(instance, self.attname)
        
        # Nothing to update if we have no file and not being forced to update.
        if not file and not force:
            return
        update_field = self.bytesize_field and not(getattr(instance, self.bytesize_field))
        
        # When the field has a value, we are most likely loading
        # data from the database or updating an image field that already had
        # an image stored.  In the first case, we don't want to update the
        # field because we are already getting the value from the
        # database.  In the second case, we do want to update the
        # field and will skip this return because force will be True since we
        # were called from ImageFileDescriptor.__set__.
        if not(update_field) and not force:
            return

        # file should be an instance of ImageFieldFile or None.
        bytesize = None
        if file:
            bytesize = file.size
            
        # ok, update
        if self.bytesize_field:
            setattr(instance, self.bytesize_field, bytesize)

    def formfield(self, **kwargs):
        # if max_len is None, the formfield is unlimited length
        max_length = None
        if (self.form_limit_filepath_length):
            max_length = self.max_length
        return super().formfield(**{
            'form_class': form_fields.TighterImageField,
            'max_length': max_length,
            'max_size': self.max_size,
            'accept_formats' : self.accept_formats,
            **kwargs,
        })




from importlib import import_module
from importlib.util import find_spec as importlib_find
from django.apps import apps

def autodiscover_modules(
        *module_names, 
        parent_modules=[], 
        find_in_apps=True, 
        not_core_apps=False
    ):
    """
    Auto-discover named modules on module paths.
    Fails silently when not present. 
    Forces an import on the module to recover any requests.
    Very similar to django.utils.module_loading.autodiscover_modules,
    but it's not.
    
    module_names 
        module names to find
    parent_modules 
        hardcoded list of module paths to search
    find_in_apps 
        seek for modules in registered apps
    not_core_apps 
        remove 'django' paths from any given list
    return
        list of modules loaded
    """
    app_modules = []
    if (find_in_apps):
        app_modules = [a.name for a in apps.get_app_configs()]
    if (not_core_apps):
        app_modules = [p for p in app_modules if (not p.startswith('django'))]
    module_parents = [*parent_modules, *app_modules]
    r = []
    for module_parent in module_parents:
        for name in module_names:
            # Attempt to import the app's module.
            try:
                p = module_parent + '.' + name
                import_module(p)
                r.append(p)
            except Exception:
                # if the module doesn't exist, ignore the error, if
                # it does and threw an error, bubble up.
                #NB: Django's module_has_submodule is not a good test. May 
                # fail as code, so module not exist.
                if (importlib_find(module_parent + '.' + name) is not None):
                    raise
    return r



from django import template
from django.utils.safestring import mark_safe
from django.forms.utils import flatatt
from django.core.exceptions import ImproperlyConfigured
from django.template.base import kwarg_re
from django.conf import settings
import os

#from django.utils.functional import cached_property
from image_lite import module_utils
from image_lite import utils

register = template.Library()

def arg_unquote(a, token, arg_name):
    if not (a[0] == a[-1] and a[0] in ('"', "'")):
       raise template.TemplateSyntaxError(
            "image tag {} argument must be in quotes. tag:{{% {} %}} value:{}".format(
            arg_name,
            token.contents,
            a
       ))
       
    return a[1:-1]

def to_kwargs(token, kwlumps):
    kwargs = {}

    # Tag is tokenizing is from strings, and that depends on 
    # quotes for whitespace and separation.
    # However, the result must be quote-stripped for flatattr() to do
    # it's Django-compliant job in ImageNode.
    # (It might be nice to do further parsing on keyword values, to 
    # allow vars and so forth. But in this atmosphere...)   
    for kw in kwlumps:
        match = kwarg_re.match(kw)
        if not match:
            raise template.TemplateSyntaxError("Malformed arguments to image tag. tag:{{% {} %}} kw:{}".format(
               token.contents,
               kw
           ))
        k, v = match.groups()
        if not (v[0] == v[-1] and v[0] in ('"', "'")):
           raise template.TemplateSyntaxError(
               "image tag keyword arguments must be in quotes. tag:{{% {} %}} kw:{}".format(
               token.contents,
               kw
           ))
        kwargs[k] = v[1:-1]

    return kwargs


class GetImageNode(template.Node):
    def __init__(self, reform_dir, file_path, kwargs):
        self.reform_dir = reform_dir
        self.file_path = file_path
        self.kwargs = kwargs

    def render(self, context):
        try:
            return mark_safe('<img src="{}" {} />'.format(
                self.file_path, 
                flatatt(self.kwargs)
                ))
        except template.VariableDoesNotExist:
            return ''
                            
@register.tag(name="image_fixed")
def image_fixed_tag(parser, token):
    '''
    Generate a URL for an image. 
    
    This tag seaches for a given filename.
    
    The image model table is not searched---the model is used to gather 
    some prperties, notably the path to image files. 
    This tag is unable to use filters, because it is not looking for
    anything.

    The tag is a full tag constructor. Keywords become tag attributes,

        {% image_url review review_images.ReviewImage class="narrow" %}
    
    model_name
        reference to a AbstractImage subclass in dotted notation. Must 
        be full path relative to the project base e.g. pages.MainImage.
    
    file_name
        name of a file, without path or extension.
        
    filter_name
        Name of a filter. No need for a module path, because the filter 
        should be located with the model. 
    
    kwargs 
        Will be added as attributes to the final tag.
    ''' 
    lumps = token.split_contents()

    if(len(lumps) < 3):
        raise template.TemplateSyntaxError(
            "Image Fixed tag needs three arguments. tag:{{% {} %}}".format(
                token.contents,
            ))
            
    tag_name = lumps[0]
    model_name = lumps[1]
    file_name = lumps[2]
    filter_name = lumps[3]
    kwargs = to_kwargs(token, lumps[4:])

    # get reform dir from the image model
    im = module_utils.get_image_model(model_name)
    reform_dir = os.path.join(settings.MEDIA_URL, im.reform_dir)
    
    # extension comes from the filter
    f = im.get_filter(filter_name)     
    extension = f.format
            
    # construct an alt
    if (not 'alt' in kwargs):
        kwargs['alt'] = f"image of {file_name}"
    full_name = file_name + '-' + filter_name  + '.' + extension
    file_path = os.path.join(reform_dir, full_name)

    return GetImageNode(reform_dir, file_path, kwargs)



class ImageNode(template.Node):
    def __init__(self, reform_dir,  template_var, filename_callable, kwargs):
        self.reform_dir = reform_dir
        self.template_var = template.Variable(template_var)
        self.filename_callable = filename_callable      
        self.kwargs = kwargs
        
    def render(self, context):
        try:
            template_var = self.template_var.resolve(context)
            
            # get the basename
            file_name = getattr(template_var, self.filename_callable)()

            # construct the alt
            if (not 'alt' in self.kwargs):
                self.kwargs['alt'] = f"image of {file_name.rsplit('-', 1)[0]}"
            filepath = os.path.join(self.reform_dir, file_name)
            return mark_safe('<img src="{}" {} />'.format(
                filepath, 
                flatatt(self.kwargs)
                ))
        except template.VariableDoesNotExist:
            return ''
              
@register.tag(name="image")
def image_tag(parser, token):
    '''
    Generate a URL for an image. 
    
    This tag generates a URL by getting an image directory path, then 
    prepending that to the return from a callable (for the filename).

    The tag is a full tag constructor. Keywords become tag attributes,

        {% image_url review review_images.ReviewImage class="narrow" %}
    
    model_name
        reference to a AbstractImage subclass in dotted notation. Must 
        be full path relative to the project base e.g. pages.MainImage.

    template_var
        A template var containing the object on which the callable will
        be called. 
        
    filename_callable
        name of a method on the repository model. The method should 
        return the filename of the image, including filter and file 
        extension (the path is constructed by tag code)
    
    kwargs 
        Will be added as attributes to the final tag.
    ''' 
    lumps = token.split_contents()

    if(len(lumps) < 3):
        raise template.TemplateSyntaxError(
            "Image tag needs three arguments. tag:{{% {} %}}".format(
                token.contents,
            ))
            
    tag_name = lumps[0]
    model_name = lumps[1]
    template_var = lumps[2]
    filename_callable = lumps[3]
    kwargs = to_kwargs(token, lumps[4:])

    # get reform dir
    im = module_utils.get_image_model(model_name)
    reform_dir = os.path.join(settings.MEDIA_URL, im.reform_dir)
        
    return ImageNode(reform_dir, template_var, filename_callable, kwargs)



"""Local image handling for Django. Unobtusive, with multiple repositories, powerful filter system and scaleable data."""
__version__ = '0.0.1'

from image_lite.decorators import register
from image_lite.filters import Filter
from image_lite.registry import registry
from image_lite.module_loading import autodiscover_modules
from image_lite.model_mixins import ModelUniqueFilenameMixin


# This is placed here as a convenient point, not because it runs on 
# init.
# autodiscover_modules does a hairy import, which will 
# fail on Django initialisation, so this method is is run from 
# ImageConfig in apps.py
def autodiscover():
    autodiscover_modules(
        'image_filters', 
        parent_modules = [], 
        find_in_apps = True, 
        not_core_apps = True
    )



from django.core.management.base import BaseCommand, CommandError
from image_lite.management.commands import common



class Command(BaseCommand):
    help = 'Automatically/bulk delete images'
    output_transaction = True

    def add_arguments(self, parser):
        common.add_model_argument(parser)
        common.add_contains_argument(parser)
        parser.add_argument(
            '--bulk',
            action='store_true',
            help='Delete images by SQL (fast, but less robust)',
        )

    def handle(self, *args, **options):
        Model = common.get_image_model(options)
        qs = Model.objects.all()
        qs = common.filter_query_contains(options, qs)
        
        if options['bulk']:
            r = qs.delete()            
            if (options['verbosity'] > 0):
                print("{} image(s) deleted".format(r[0])) 
        else:
            count = 0
            for im in qs:
                r = im.delete()
                
                # count from return. May have failed!
                count += r[0]

            if (options['verbosity'] > 0):
                print("{} image(s) deleted".format(count)) 



from django.forms.fields import FileField
from django.core.exceptions import ValidationError
from django.forms.widgets import FileInput
from image_lite import validators


class TighterImageField(FileField):
    '''
    A FileField with extended validation.
    It can skip length verification, and prechecks filesize and 
    extensions.
    '''
    #NB I do't see why Imagefield thinks to_python() is the place to
    # do validation. Maybe I'm missing something, but ugty, 
    # and stripped out. 
    # The other and better reason for this existing is to get the 
    # validaors out into the forms fields, where Django feels they 
    # belong.
    def __init__(self, *,
        max_size=None,
        accept_formats=None,
        **kwargs
    ):
        self.max_size = max_size
        self.accept_formats = accept_formats
        super().__init__(**kwargs)
        
    def to_python(self, data):
        """
        Check that the file-upload field data contains a valid image (GIF, JPG,
        PNG, etc. -- whatever Pillow supports).
        """
        f = super().to_python(data)
        if f is None:
            return None
        if hasattr(f, 'seek') and callable(f.seek):
            f.seek(0)
        return f

    def validate(self, value):
        super().validate(value)
        if (self.max_size):
            validators.validate_file_size(value, self.max_size)
        validators.ImageFileExtensionValidator(
            allowed_extensions=self.accept_formats
        )(value)

    def widget_attrs(self, widget):
        attrs = super().widget_attrs(widget)
        if isinstance(widget, FileInput) and 'accept' not in widget.attrs:
            attrs.setdefault('accept', 'image/*')
        return attrs



def register():
    """
    Register the given filter classes with the image app:

    @register()
    class Thumb(filters_pillow.ResizeSmart):
        width=64
        height=64
    """
    from image_lite import Filter, registry

    def _filter_wrapper(filter_class):

        if not issubclass(filter_class, Filter):
            raise ValueError('Wrapped class must subclass image.Filter.')
            
        registry.register(filter_class)

        return filter_class

    return _filter_wrapper



from PIL import Image as PILImage
from io import BytesIO
from image_lite.filters import (
    Filter,
    FormatMixin,
    ResizeCropMixin,
    ResizeCropSmartMixin,
)
from image_lite import image_ops_pillow
from image_lite.constants import FORMAT_APP_TO_UCLIB, FORMAT_UCLIB_TO_APP



class PillowProcess:
            
    def process(self, src_file, model_args):
        src_image = PILImage.open(src_file)
        
        # write_attrs currently {format, jpeg_quality}
        # write_attrs = reform_save_info(
                    # self,
                    # FORMAT_UCLIB_TO_APP[src_image.format],
                    # model_args,
                    # )

        # stash the internal app format
        app_format = self.format
                            
        write_attrs = {'format': app_format}
        if app_format == 'jpg':
            write_attrs['jpeg_quality'] = self.jpeg_quality

        # mods on the save data
        # convert the returned format to PIL
        write_attrs['format'] = FORMAT_APP_TO_UCLIB[write_attrs['format']]
        
        # Add some general write attributes. Pillow ignores the unusable.        
        write_attrs['progressive'] = True  
        write_attrs['optimize'] = True

        # break out processing, it's the only action that changes
        # across different filters
        pil_dst = self.modify(src_image) or src_image

        # PIL raises error if given RGBA for JPEG output
        # in that case, no transparency
        if ((pil_dst.mode in ('RGBA', 'LA')) and (app_format == 'jpg')):            
            # If expressed, respect background colour. May be backing
            # a genuine transparency, so path of least surprise.
            if (hasattr(self, 'fill_color')):
                fc = self.fill_color
            else:
                fc = 'white'
            bg = PILImage.new('RGB', pil_dst.size, fc)
            bg.paste(pil_dst, pil_dst.getchannel('A'))
            pil_dst = bg
        
        #! test. I think that's the name.
        if ('jpeg_quality' in write_attrs):            
            write_attrs['quality'] = write_attrs['jpeg_quality']
        out_buff = BytesIO()
        pil_dst.save(
            out_buff,
            **write_attrs
        )
        return (out_buff, app_format)

    def modify(self, lib_image):
        '''
        Modify the image.
        'process hadles the image load and save. 'process' calls this 
        method to transform the image. Basely this is the only method
        that needs to be changed, even for a filter that performs a 
        new kind of transformation (no need to change the load and save 
        in 'process).
        Enabled methods should usually call super() to enable inherited 
        code.
        
        @lib_image the class used by the library to wrap image data
        '''
        # simple return. PIL uses returns
        return lib_image
        
    
        
class Format(FormatMixin, PillowProcess, Filter):
    '''Establish the format for an image. 
    Set iformat=None means the image is unchanged.
    '''
    
    
    
class Resize(ResizeCropMixin, Format):
    '''Resize n image.
    Shrinks inside the box defined by the given args. So the result will
    usually be smaller width or height than the given box.
    A base class e.g.
        class Large(image.Resize):
            width=513
            height=760    
            iformat='png'
    '''
    def modify(self, lib_image):
        i = image_ops_pillow.resize_aspect(
            lib_image, 
            self.width,
            self.height
        )
        return super().modify(i)

        

class Crop(ResizeCropMixin, Format):
    '''Resize and maybe re-format an image.
    A base class e.g.
        class Large(image.Crop):
            width=513
            height=760 
            tpe='png'
    '''
    def modify(self, lib_image):
        i = image_ops_pillow.crop(
            lib_image,
            self.width,
            self.height
        )
        return super().modify(i)

                
                                    
class ResizeSmart(ResizeCropSmartMixin, Format):
    '''Resize an image.
    This resize lays the image on a background of ''fill-color'.
    So the result always matches the given sizes.
    A base class e.g.
        class Large(image.ResizeSmart):
            width=513
            height=760    
            tpe='png'
            fill_color="dark-green"
    '''
    def modify(self, lib_image):
        i = image_ops_pillow.resize_aspect(
            lib_image, 
            self.width, 
            self.height
            )
        i = super().modify(i)
        i = image_ops_pillow.fill(
            i, 
            self.width, 
            self.height, 
            self.fill_color
            )
        return i




class CropSmart(ResizeCropSmartMixin, Format):
    '''Resize and maybe re-format an image.
    A base class e.g.
        class Large(image.CropSmart):
            width=513
            height=760    
            tpe='png'
            fill_color="coral"
    '''
    def modify(self, lib_image):
        i = image_ops_pillow.crop(
            lib_image, 
            self.width, 
            self.height
            )
        i = super().modify(i)
        i = image_ops_pillow.fill(
            i, 
            self.width, 
            self.height, 
            self.fill_color
            )        
        return i



from wand.image import WandImage
from io import BytesIO
from image_lite.filters import (
    Filter,
    FormatMixin,
    PhotoFXMixin,
    ResizeCropMixin,
    ResizeCropSmartMixin,
)
from image_lite import image_ops_wand
from image_lite.constants import FORMAT_APP_TO_UCLIB, FORMAT_UCLIB_TO_APP
from image_lite.decisions import reform_save_info



class WandProcess():

    def process(self, src_file, model_args):
        image = WandImage(file=src_file)
 
        # write_attrs currently {format, jpeg_quality}
        write_attrs = reform_save_info(
                    self,
                    FORMAT_UCLIB_TO_APP[image.format],
                    model_args,
                    )
 
        # break out processing, it's the only action that changes
        # across different filters
        wand_dst = self.modify(image) or image
        
        # set the format
        wand_dst.format = FORMAT_APP_TO_UCLIB[write_attrs['format']]
        
        # set JPEG and others quality
        wand_dst.compression_quality = write_attrs['jpeg_quality']
        
        # Defend against rendering jpeg transparency
        if ((wand_dst.alpha_channel) and (write_attrs['format'] == 'jpg')): 
            # Here's where Wand plays for us. Alpha channel or not, it 
            # sets a background_color. And the default is 'white', like 
            # us. So all we have to do is remove the alpha channel, 
            # because Wand doesn't like one on jpeg rendering, and Wand 
            # will compose down.
            wand_dst.alpha_channel = 'remove'
        out_buff = BytesIO()

        wand_dst.save(
            out_buff,
        )
        
        return (out_buff,  write_attrs['format'])

    def modify(self, lib_image):
        '''
        Modify the image.
        'process hadles the image load and save. 'process' calls this 
        method to transform the image. Basely this is the only method
        that needs to be changed, even for a filter that performs a 
        new kind of transformation (no need to change the load and save 
        in 'process).
        Enabled methods should usually call super() to enable inherited 
        code.
        
        @lib_image the class used by the library to wrap image data
        '''
        # Wand often alters in-place, but generates images for some 
        # processing.
        return lib_image



class Format(PhotoFXMixin, FormatMixin, WandProcess, Filter):
    '''Establish the format for an image. 
    Set iformat=None means the image is unchanged.
    '''
    def modify(self, lib_image):
        return image_ops_wand.photoFX(
            lib_image, 
            self.pop, 
            self.greyscale, 
            self.warm, 
            self.night, 
            self.strong, 
            self.no,
            self.watermark,
        )    


                        
class Resize(ResizeCropMixin,  Format):
    '''Resize n image.
    Shrinks inside the box defined by the given args. So the result will
    usually be smaller width or height than the given box.
    A base class e.g.
        class Large(image.Resize):
            width=513
            height=760    
            iformat='png'
    '''
    def modify(self, lib_image):
        i = image_ops_wand.resize_aspect(
            lib_image,
            self.width,
            self.height
            )
        return super().modify(i)


        
class Crop(ResizeCropMixin, Format):
    '''Resize and maybe re-format an image.
    A base class e.g.
        class Large(image.Crop):
            width=513
            height=760    
            tpe='png'
    '''
    def modify(self, lib_image):
        i = image_ops_wand.crop(
            lib_image,
            self.width,
            self.height
        )
        return super().modify(i)



class ResizeSmart(ResizeCropSmartMixin, Format):
    '''Resize an image.
    This resize lays the image on a background of ''fill-color'.
    So the result always matches the given sizes.
    A base class e.g.
        class Large(image.ResizeSmart):
            width=513
            height=760    
            tpe='png'
            fill_color="dark-green"
    '''
    def modify(self, lib_image):
        i = image_ops_wand.resize_aspect(
            lib_image, 
            self.width, 
            self.height
            )
        i = super().modify(i)
        i = image_ops_wand.fill(
            lib_image, 
            self.width, 
            self.height, 
            self.fill_color
            )
        return i



class CropSmart(ResizeCropSmartMixin, Format):
    '''Resize and maybe re-format an image.
    A base class e.g.
        class Large(image.CropSmart):
            width=513
            height=760    
            tpe='png'
            fill_color="coral"
    '''
    def modify(self, lib_image):
        i = image_ops_wand.crop(
            lib_image, 
            self.width, 
            self.height
            )
        i = super().modify(i)
        i = image_ops_wand.fill(
            lib_image, 
            self.width, 
            self.height, 
            self.fill_color
            )        
        return i



import zlib
import struct
from pathlib import Path
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from django.template.defaultfilters import filesizeformat
from django.utils.deconstruct import deconstructible
from image_lite.constants import (
    IMAGE_FORMATS,
    EXTENSION_TO_APP_FORMAT,
    FORMAT_UCLIB_TO_APP
)

def get_pillow_attribute(file, pil_attr):
    """
    Returns an attribute from a Pillow image of the given file,
    file
        a file. wrappable by Pillow. Must be open
    pil_attr
        the attribute to read.
    return 
        the attribut, or None, if unreadaable. the file is not closed.
    """
    # This devilish, ticketed code is lifted from 
    # django.core.files.images, where it seeks image dimensions.
    # Unfortunately, that code is not configurable nor overridable.
    from PIL import ImageFile as PillowImageFile

    p = PillowImageFile.Parser()
    file_pos = file.tell()
    file.seek(0)
    try:
        # Most of the time Pillow only needs a small chunk to parse the
        # image and get data, but with some TIFF files Pillow needs to
        # parse the whole file.
        chunk_size = 1024
        while 1:
            data = file.read(chunk_size)
            if not data:
                break
            try:
                p.feed(data)
            except zlib.error as e:
                # ignore zlib complaining on truncated stream, just feed more
                # data to parser (ticket #19457).
                if e.args[0].startswith("Error -5"):
                    pass
                else:
                    raise
            except struct.error:
                # Ignore PIL failing on a too short buffer when reads return
                # less bytes than expected. Skip and feed more data to the
                # parser (ticket #24544).
                pass
            except RuntimeError:
                # e.g. "RuntimeError: could not create decoder object" for
                # WebP files. A different chunk_size may work.
                pass
            if p.image:
                return getattr(p.image, pil_attr)
            chunk_size *= 2
        return None
    finally:
        file.seek(file_pos)
            
    
@deconstructible
#class ImageFileDataConsistencyValidator:
class ImageFileExtensionValidator:
    '''
    Check the extension of an image file.
    Checks the extension against the internally defined set of file 
    extensions. Then does a Pillow read, to see what Pillow thinks,
    and compares the two.

    allowed_extensions
        a lower case, short format, list
    '''
    messages = {
        'no_extension': _(
            "No file extension. Allowed extensions: %(extensions)s."
        ),
        'extension_not_allowed': _(
            "File extension %(extension)s not allowed. allowed extensions: %(extensions)s."
        ),
        'format_not_readable': _(
            "File not readable as an image."
        ),
        'format_extension_mismatch': _(
            "Found format does not match extension."
        ),
    }
    allowed_extensions = IMAGE_FORMATS
    ext2app = EXTENSION_TO_APP_FORMAT
    pil2app = FORMAT_UCLIB_TO_APP
    
    def __init__(self, allowed_extensions=None):
        if allowed_extensions is not None:
            self.allowed_extensions = allowed_extensions
        self.allowed_extensions_message = ', '.join(self.allowed_extensions)
        
    def __call__(self, v):
        extension = Path(v.name).suffix[1:].lower()
        if (len(extension) < 2):
            raise ValidationError(
                    self.messages['no_extension'],
                    code='no_extension',
                    params={'extensions': self.allowed_extensions_message},
                )
        if (not(extension in self.allowed_extensions)):
            raise ValidationError(
                    self.messages['extension_not_allowed'],
                    code='extension_not_allowed',
                    params={'extension': extension, 'extensions': self.allowed_extensions_message},
                )
        try:
            raw_read_format = get_pillow_attribute(v.file, 'format')
        except Exception as e:
            raise ValidationError(
                    self.messages['format_not_readable'],
                    code='format_not_readable',
                ) from e
        read_format = self.pil2app[raw_read_format]
        declared_format = self.ext2app[extension]
        if (read_format != declared_format):
            raise ValidationError(
                    self.messages['format_extension_mismatch'],
                    code='format_extension_mismatch',
                )
        
    def __eq__(self, other):
        return (
            isinstance(other, self.__class__) and
            self.allowed_extensions == other.allowed_extensions
        )

def validate_image_file_consistency(value):
    return ImageFileExtensionValidator()(value)

def validate_file_size(file, max_size):
    '''
    Check size of a file.
    Must be protected from out of range values like None.
    
    max_size
        in bytes
    file
        a Django file class. The file must be open.
    '''
    # Check filesize
    if (file.size > max_size):
        raise ValidationError(
            "This file is too big* ({}). Maximum filesize {}.".format(
                filesizeformat(file.size),
                filesizeformat(max_size)
            ))



from django.core.management.base import BaseCommand
from pathlib import Path
from image_lite.management.commands import common



class Command(BaseCommand):
    help = 'Resync local image files to the database. Default is to print status.'
    output_transaction = True
    
    def add_arguments(self, parser):
        common.add_model_argument(parser)
        parser.add_argument(
            '--remove-orphaned-files',
            action='store_true',
            help='Delete orphaned files',
        )
        parser.add_argument(
            '-a',
            '--add-orphaned-files',
            action='store_true',
            help='Add models for orphaned files (model attributes default)',
        )
        parser.add_argument(
            '--remove-orphaned-models',
            action='store_true',
            help='Delete orphaned models (with no matching file)',
        )        
        
    def originals_fp_list(self, file_dir):
        return [f for f in file_dir.iterdir() if f.is_file()]

    def db_image_fp_list(self, Model, storage_location):
        ''' 
        Read all image model paths
        return 
            full-path Paths
        '''
        val_list = Model.objects.values_list('src', flat=True)
        return [storage_location / rel_path for rel_path in val_list]
            
    def db_image_list(self, Model, storage_location):
        ''' 
        Read all image models for pk and src
        return
            list -> dict {fieldname: fieldvalue} supplemented with 'full_path' key/value
        '''
        val_list = Model.objects.values('pk', 'src')
        for e in val_list:
            e['full_path']  = storage_location / e['src']
        return  val_list  

    def db_image_count(self, Model):
        return Model.objects.count()
        
    def handle(self, *args, **options):        
        Model = common.get_image_model(options)
        
        # get the model dir path
        # Mad Django
        storage = Model._meta.get_field('src').storage 
          
        # location is MEDIA_ROOT for local files
        location = Path(storage.location)
        
        # path is a path constructor from MEDIA_ROOT
        file_dir = Path(storage.path(Model.upload_dir))    
        fl = self.originals_fp_list(file_dir)
        
        if options['remove_orphaned_models']:            
            dbl = self.db_image_list(Model, location)
            model_no_file = [m for m in dbl if not(m['full_path'] in fl)]           
            pks = [m['pk'] for m in model_no_file]
            r = Model.objects.filter(pk__in=pks).delete()
            if (options['verbosity'] > 0):
                print("{} model(s) deleted".format(r[0])) 
            if (options['verbosity'] > 2):
                for m in model_no_file:
                    print(m['src'])
                    
        elif options['remove_orphaned_files']:
            count = 0
            fail = []
            dbl = self.db_image_fp_list(Model, location)
            file_no_model = set(fl) - set(dbl)
            for path in file_no_model:
                try:
                    path.unlink()
                    count += 1
                except Exception:
                    fail.append(path.stem)
            if (options['verbosity'] > 0):
                print("{} image(s) deleted".format(count)) 
                if (len(fail) > 0):
                    print("{} image(s) failed to delete. Path: '{}'".format(
                        len(fail),
                        "', '".join(fail),
                        ))
        elif options['add_orphaned_files']:
            # Add models for orphaned files
            count = 0
            fail = []
            dbl = self.db_image_fp_list(Model, location)
            file_no_model = set(fl) - set(dbl)
            for path in file_no_model:
                # Make new models.
                # Worth explaining:
                # How do we register orphaned files?
                # When the model is saved, the generated FieldFile will 
                # not recognise the file is stored and attempt a copy 
                # onto itself. This should fail silently, in linux, 
                # but Django throws a FileExistsError, then makes a 
                # renamed copy.
                # Preferebly, we would tell FieldFile the storage is 
                # committed. Interesingly, a simple string path, as 
                # opposed to an open File or Imagefile, assumes the
                # image is commited.
                #
                # The update effect in Imagefields will not work with 
                # Paths, only strings. Also, it works relative to /media
                # (so path relativization here)
                #
                # Presuming all is default in the new model. If you need
                # to use this command, sure you can work from there.
                i = Model(
                    src = str(Path(Model.upload_dir) / path.name),
                ) 
                try:
                    i.save()
                    count += 1
                except Exception:
                    fail.append(path.stem)
        
            if (options['verbosity'] > 0):
                print("{} image(s) created".format(count)) 
                if (len(fail) > 0):
                    print("{} image(s) failed import. Path: '{}'".format(
                        len(fail),
                        "', '".join(fail),
                        )) 
        else:
            # status report
            if (options['verbosity'] > 0):
                header = "SYNC STATUS: {}".format(Model.__name__)
                print("-" * (len(header) + 1))
                print(header)
                print("-" * (len(header) + 1))
            model_count = self.db_image_count(Model)
            file_count = len(fl)
            if ( (model_count != file_count) or options['verbosity'] > 0):
                print("model_count: {}".format(model_count))
                print("file_count: {}".format(file_count))
            if (model_count > file_count):
                print('[warning] files appear to be missing. Try --remove-orphaned-models')
            elif (file_count > model_count):
                print('[warning] models appear to be missing. Try --add-orphaned-files or --remove-orphaned-files')
            else:
                print('** No sync issues detected')
            
            



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['image_lite', 'image_lite.migrations', 'image_lite.templatetags']

package_data = \
{'': ['*'],
 'image_lite': ['management/*',
                'management/commands/*',
                'templates/*',
                'templates/image/*']}

setup(name='django_image_lite',
      version='0.0.1',
      description='Local image handling for Django. Unobtusive, with multiple repositories, powerful filter system and scaleable data.',
      author=None,
      author_email='robert crowther <rw.crowther@gmail.com>',
      url=None,
      packages=packages,
      package_data=package_data,
     )



import datetime

from django.core.management.base import BaseCommand, CommandError
from image_lite.management.commands import common
from pathlib import Path



class Command(BaseCommand):
    help = 'List images'
    output_transaction = True
    
    def add_arguments(self, parser):
        common.add_model_argument(parser)
        common.add_contains_argument(parser)
        parser.add_argument(
            '-w',
            '--weeks-back',
            type=int,
            help='Search for entries from now to "count" weeks back',
        )
                
    def handle(self, *args, **options):
        Model = common.get_image_model(options)
        qs = Model.objects
        qs = common.filter_query_contains(options, qs)
        if (options["weeks_back"]):
            week_num_now =  datetime.date.isocalendar(datetime.date.today())[1]
            
            if (options["weeks_back"] > week_num_now or options["weeks_back"] < 0):
                raise CommandError("Value of --weeks_back option cant be negative or larger than the year. Today is in week {}".format(week_num_now))             
            week_num_back = week_num_now - options["weeks_back"]
            qs = qs.filter(upload_date__week__gte=week_num_back)

        for e in qs.all():
            print("{} {}".format(e.pk, Path(e.src.name).stem))



from image_lite.utils import url_absolute_static_aware
from wand.image import WandImage
from wand.color import Color
from wand.drawing import Drawing
import math


def photoFX(wand, pop, grayscale, warm, night, strong, no, watermark):
    if (pop):
        wand.level(0.2, 0.9, gamma=1.1)

    if (grayscale):
        # Reputedly fast and preserves transparency. Small errors 
        # will creep in, the visual effect is what matters.
        matrix = [[0.333, 0.333, 0.333],
          [0.333, 0.333, 0.333],
          [0.333, 0.333, 0.333]]
        wand.color_matrix(matrix)
        
    if (night):
        matrix = [[0.2, 0.0, 0],
           [0, 0.3, 0],
           [0, 0, 1]]
        wand.color_matrix(matrix)
        
    if (warm):
        # Yes, there are problems with HSB, but it will be ok for this.
        wand.modulate(brightness=100.0, saturation=100.0, hue=94.0)

    if (strong):
        wand.modulate(brightness=100.0, saturation=160.0, hue=100.0)

    if (no):
        insetX = wand.width >> 4
        insetY = wand.height >> 4
        insetX1 = wand.width - insetX
        insetY1 = wand.height - insetY
        with Drawing() as draw: 
            draw.stroke_color = Color('red')
            draw.stroke_width = 12
            draw.line((insetX, insetY1), (insetX1, insetY))
            draw.line((insetX, insetY), (insetX1, insetY1))        
            draw(wand)
            
    if (watermark):
        url = url_absolute_static_aware(watermark)
        scale_width = wand.width >> 1
        
        with WandImage(filename=watermark) as overlay:
            # scale the overlay to the inset.            
            overlay.resize(scale_width, round(scale_width/overlay.width * overlay.height))
            wand.composite(
                overlay, 
                left=None, 
                top=None, 
                operator='dissolve', 
                arguments='35', 
                gravity='center'
            )
    return wand


def resize_aspect(wand, width, height):
    '''
    Resize if the image is too big, preserving aspect ratio.
    Will not resize if dimensions match or are less than the source.
    @return an image within the given dimensions. The image is usually 
    smaller in one dimension than the given space.
    '''
    width_reduce = wand.width - width
    height_reduce = wand.height - height
  
    if (width_reduce > height_reduce and width_reduce > 0):
        h = math.floor((width * wand.height)/wand.width)
        wand.resize(width, h)        

    #NB the equality. On the not unlikely chance that the width 
    # reduction is the same as the height reduction (for example, 
    # squares), reduce by height.
    elif (height_reduce >= width_reduce and height_reduce > 0):
        w = math.floor((height * wand.width)/wand.height)
        wand.resize(w, height)
    return wand



def crop(wand, width, height):
    '''
    Crop if image is too big.
    Crop is anchored at centre. 
    Will not crop if both dimensions match or are less than the source.
    @return an image within the given dimensions. The image may be 
    smaller in one dimension than the given space.
    '''
    current_width = wand.width
    current_height = wand.height
    
    width_reduce = current_width - width
    height_reduce = current_height - height

    # Only crop if the image is too big
    if ((width_reduce > 0) or (height_reduce > 0)):
        x = 0
        y = 0
        if (width_reduce > 0):
            x = width_reduce >> 1
        if (height_reduce > 0):
            y = height_reduce >> 1
        # Crop!
        wand.crop(left=x, top=y, width=width, height=height)
    return wand


def fill(wand, width, height, fill_color="white"):
    '''
    Fill round an image to a box.
    Image must be smaller than the giveen box. Checking is 
    regarded as a seperate operation.
    '''
    current_width = wand.width
    current_height = wand.height

    with Color(fill_color) as bg:
        wand.background_color = bg
        x = (width - current_width) >> 1
        y = (height - current_height) >> 1

        # the negative is deliberate. Wand I think is anchored to *bottom* left.
        wand.extent(width=width, height=height, x=x, y=-y)

    return wand



from django.core.exceptions import ImproperlyConfigured
from django.core.management.base import CommandError
from image_lite import module_utils
from pathlib import Path


def add_model_argument(parser):
        parser.add_argument(
            'model',
            type=str,
            help='Target a model derived from AbstractImage, form:is <app.model>.',
        )
        
def get_image_model(options):
    model_path = options['model']        
    try:
        Image = module_utils.get_image_model(model_path)
    except ImproperlyConfigured as e:

        # Stock exception system not working for this
        raise  CommandError(e.args[0])
    return Image

def get_storage(image_class_or_instance):
    return image_class_or_instance._meta.get_field('src').storage 

def get_reform_path(options):
    im = get_image_model(options)
    #! use im.reform_dir_path()
    storage = im._meta.get_field('src').storage   
    path = Path(storage.location) / im.reform_dir
    if (not(path.is_dir())):
        raise CommandError(f"Reform path not recognised as a directory. path: '{path}'")
    return path 
            
def add_contains_argument(parser):
    parser.add_argument(
        '-c',
        '--contains',
        type=str,
        help='Search for src-names containing this text',
    )
    
def filter_query_contains(options, queryset):
    if (options["contains"]):
        queryset = queryset.filter(src__icontains=options["contains"])
    return queryset



from pathlib import Path
import os
from django.core.management.base import BaseCommand, CommandError
from django.core.files.images import ImageFile
from image_lite.management.commands import common



class Command(BaseCommand):
    help = 'Automatically/bulk add images to an image model. The command tries to ignore errors and continue. It will append to existing collections. Attributes other than file data will default.'

    def add_arguments(self, parser):
        common.add_model_argument(parser)
        parser.add_argument(
            'src_path', 
            type=str
        )

                         
    def handle(self, *args, **options):
        Model = common.get_image_model(options)
            
        # get the path of src dir
        src_path = options['src_path']
        src_dir = Path(src_path)
        if (not(src_dir.is_dir())):
            raise CommandError("Source provided is not recognised as a directory. path: '{}'".format(src_path))
        
        # get full paths of files in the src dir
        src_filepaths = [f.resolve() for f in src_dir.iterdir() if f.is_file()]
       
        # build models and save
        # I think we can not use bulk_create because we want to be 
        # cautious and succeed as much as possible
        count = 0
        fail = []
        for path in src_filepaths:
            with open(path, 'rb') as f:                
                basename = os.path.basename(path)
                #media_relative_path = os.path.join(root, basename) 
                i = Model(
                    src = ImageFile(f, name=basename),
                ) 
                try:
                    i.save()
                    count += 1
                except Exception:
                    fail.append(basename)
        
        # output some results            
        if (options['verbosity'] > 0):
            print("{} image(s) created".format(count)) 
            if (len(fail) > 0):
                print("{} image(s) failed save. Basenames: '{}'".format(
                    len(fail),
                    "', '".join(fail),
                    )) 



from django.core.management.base import BaseCommand, CommandError
from image_lite.management.commands import common
from pathlib import Path



class Command(BaseCommand):
    help = 'Automatically/bulk delete reform images'
    output_transaction = True

    def add_arguments(self, parser):
        common.add_model_argument(parser)
        common.add_contains_argument(parser)
        
    def handle(self, *args, **options):
        reform_path = common.get_reform_path(options)
        reform_list = [fp for fp in reform_path.iterdir() if fp.is_file()]
        contains_test = options["contains"]
        if (contains_test):
            reform_list = filter(lambda fn: contains_test in str(fn), reform_list)
        count = 0
        for fp in reform_list:
            try:
                Path(fp).unlink(missing_ok=False)
                count += 1
            except FileNotFoundError:
                # ignore, and carry on
                pass
        if (options['verbosity'] > 0):
            print("{} reforms deleted".format(count)) 
